' flashback.bas version 1.0.00 - a fun and nostalgic terminal program using the esp8266 library
' Written by Rich Martin (datawiz) August 2020
'
option explicit
option default none
option console screen

#include "esp8266var.inc"

dim SERVER_IP$  = ""
dim SERVER_PORT$= ""
dim COM_PORT$ = "COM2"
dim COM_SPEED$ = "115200"
dim COM_BUFFER$ = "256"
dim C$ = ""
dim RECVCLICK% = 4800
dim RECVCLICKDELAY1% = 1
dim RECVCLICKDELAY2% = 3

const kreprate = 200
dim kf%, kl%, kp%
dim altflag%, winflag%
dim debug% = 0

const xoffset% = 10
const yoffset% = 10
const width% = 80
const height% = 24

dim onlineflag% = 0
dim cursortype% = 0
dim blinkflag% = 0
dim soundflag% = 1
dim scr_text$(width,height)
dim scr_color%(width,height)
dim fwidth% = mm.info(fontwidth)
dim fheight% = mm.info(fontheight)
dim xpos% = 0
dim ypos% = 0
dim in$, k$

color rgb(green), rgb(black)
cls
setbuffer
' screen box
rbox fwidth%*(xoffset%-1), fheight%*(yoffset%-1), (width+2)*fwidth%, (height+2)*fheight%, 8, rgb(green), 0
' status/input box
box fwidth%*(xoffset%-1), (height%+12)*fheight%, (width+2)*fwidth%, 5*fheight%, 1, rgb(green), 0

'clearscreen
' open com port to esp8266
esp.var.com.filenum% = 6
open COM_PORT$+":"+COM_SPEED$+","+COM_BUFFER$+",esp.input.loop" as esp.var.com.filenum%

' tell esp.input.loop to not output recieved data from esp
esp.var.echo% = 0
' in case we are in transparent mode, escape out and close connection
esp.wifi.escape("AT+CIPCLOSE")
esp.wifi.restart

' set esp to echo back our input
esp.echo.set("OFF")
' set esp.input.loop to output received data from esp
esp.var.echo% = 1
' set esp.input.loop mode to normal output (output data received byte by byte)
esp.var.waitfor.mode% = esp.waitfor.mode.normal

onlineflag% = 0
updateonline
printstatus("> ALT-? for Help")

' handle input
do
  if blinkflag% = 0 then blinkcursor
  C$ = getchar$()
  if altflag% = 1 then
  
    select case lcase$(C$)
      case "c"
        if winflag% = 1 then
          cursortype% = cursortype% + 1
          if cursortype% > 3 then cursortype% = 0
          printstatus("> Cursor Type "+str$(cursortype%))
        else
          if onlineflag% = 0 then 
            dialogConnect
          else
            printstatus("> Already online. Disconnect first.")            
          end if
        end if
      case "d"
        if winflag% = 1 then
          if debug% = 0 then 
            printstatus("> Debug Mode On")
            debug% = 1
          else
            printstatus("> Debug Mode Off")
            debug% = 0
          end if
        end if
      case "x"
        esp.var.echo% = 0
        esp.wifi.escape("AT+CIPCLOSE")
        pause 1000
        printstatus("> Disconnected")
        esp.var.echo% = 1
        onlineflag% = 0
        updateonline
      case "q"
        printstatus("> Quit")
        goto QUIT
      case "r"
        printstatus("> Restarting esp8266")
        esp.wifi.restart
      case "s"
        if soundflag% = 0 then
          printstatus("> Sound On")
          soundflag% = 1
        else
          printstatus("> Sound Off")
          soundflag% = 0
        end if
      case "h"
          printstatus("> Home Screen")
          clearscreen
      case "e"    
        if esp.var.echo% = 0 then
          printstatus("> Local Echo On")
          esp.var.echo% = 1
        else
          printstatus("> Local Echo Off")
          esp.var.echo% = 0
        end if  
      case "v"
        if winflag% = 1 then
          printstatus("> Version 1.0.0 - by Rich Martin")
        end if
      case "?"
        dialogHelp  
    end select
  else
    if C$ <> "" then
      if debug% then printstatus("> DEBUG OUT HEX: 0x"+hex$(asc(C$),2))
      if C$ = chr$(10) then print #esp.var.com.filenum%, chr$(13);
      print #esp.var.com.filenum%, C$;
      if onlineflag% = 0 then 
        if C$ = chr$(10) then printc(chr$(13))
        printc(C$)
      end if        
      if soundflag% = 1 then play tone 1600, 1600, 10
    end if  
  end if
loop

QUIT:
esp.var.echo% = 0
esp.wifi.escape("AT+CIPCLOSE")
pause 1000
close esp.var.com.filenum%
cls
end

#include "flashback-esplib.inc"

sub updateonline
  if onlineflag% = 0 then print @(fwidth%*(xoffset%+74)-2,(height%+12)*fheight%+2,2) "OFFLINE"
  if onlineflag% = 1 then print @(fwidth%*(xoffset%+74)-2,(height%+12)*fheight%+2,2) "ON LINE"
end sub

sub dialogHelp
  local C$
  const ox = 33
  const oy = 15
  blit read 2, ox*fwidth%, oy*fheight%, 32*fwidth%, 13*fheight%
  box ox*fwidth%, oy*fheight%, 32*fwidth%, 13*fheight%, 1,,rgb(black)
  print @((ox+2)*fwidth%,(oy+1)*fheight%) "ALT-C Connect to a Server";
  print @((ox+2)*fwidth%,(oy+2)*fheight%) "ALT-X Disconnect from Server";
  print @((ox+2)*fwidth%,(oy+3)*fheight%) "ALT-S Toggle sound on/off";
  print @((ox+2)*fwidth%,(oy+4)*fheight%) "ALT-E Toggle echo on/off";
  print @((ox+2)*fwidth%,(oy+5)*fheight%) "ALT-H Home/Clear the screen";
  print @((ox+2)*fwidth%,(oy+6)*fheight%) "ALT-Q Quit and exit Terminal";
  print @((ox+2)*fwidth%,(oy+7)*fheight%) "ALT-R Restart the esp8266";
  print @((ox+2)*fwidth%,(oy+8)*fheight%) "";
  print @((ox+2)*fwidth%,(oy+9)*fheight%) "ALT+WIN-C Toggle cursor mode";
  print @((ox+2)*fwidth%,(oy+10)*fheight%) "ALT+WIN-D Toggle debug on/off";
  print @((ox+2)*fwidth%,(oy+11)*fheight%)"ALT+WIN-V Show Version info";
  do while inkey$ = "" : loop
  box ox*fwidth%, oy*fheight%, 32*fwidth%, 13*fheight%,1,rgb(black),rgb(black)
  blit write 2, ox*fwidth%, oy*fheight%
  do while inkey$ = "" : loop
end sub

sub dialogConnect
  local C$
  const ox = 15
  const oy = 20
  blit read 1, ox*fwidth%, oy*fheight%, 70*fwidth%, 3*fheight%
  box ox*fwidth%, oy*fheight%, 70*fwidth%, 3*fheight%, 1,,rgb(black)
  print @((ox+1)*fwidth%,(oy+1)*fheight%)"";
  input "Connect> Enter Server, Port: ", SERVER_IP$, SERVER_PORT$;
  box ox*fwidth%, oy*fheight%, 70*fwidth%, 3*fheight%,1,rgb(black),rgb(black)
  blit write 1, ox*fwidth%, oy*fheight%
  do while inkey$ = "" : loop
  if SERVER_IP$ <> "" and SERVER_PORT$ <> "" then
    printstatus("> Connecting to "+SERVER_IP$+" port "+SERVER_PORT$)
    if soundflag% = 1 then SoundsConnect
    esp.var.echo% = 0
    esp.tcp.client.connect(SERVER_IP$, SERVER_PORT$)
    esp.var.echo% = 1
    if esp.var.return$ = "" then 
      printstatus("> Connected")
      onlineflag% = 1
      updateonline
    else
      printstatus("> Connect Failed")
    end if
  end if
end sub

sub SoundsConnect
  local DTMF%(10,2)
  local C%, D%, N%, Y%
  local CH$, PHONE$
  data 941,1477,697,1209,697,1336,697,1447,770,1209,770,1336,770,1447,852,1209,852,1336,852,1447
  
  PHONE$ = SERVER_IP$
  
  restore
  for Y%=0 to 9
    read D% : DTMF%(Y%,0)=D%
    read D% : DTMF%(Y%,1)=D%
  next Y%

  play tone 445,350,2000 :  pause 2000
  for C% = 1 to LEN(PHONE$)
    CH$=MID$(PHONE$,C%,1)
    select case asc(ucase$(CH$))
      case asc("A") to asc("C")
        N%=2
      case asc("D") to asc("F")
        N%=3
      case asc("G") to asc("I")
        N%=4
      case asc("J") to asc("L")
        N%=5
      case asc("M") to asc("O")
        N%=6
      case asc("P") to asc("S")
        N%=7
      case asc("T") to asc("V")
        N%=8
      case asc("W") to asc("Z")
        N%=9
      case else
        N%=-1
    end select

    if ASC(CH$) >= 48 and ASC(CH$) <=57 then  ' char is between 0-9
      N%=VAL(CH$)  
    endif

    if N%>=0 then
      play tone DTMF%(N%,0), DTMF%(N%,1), 100
      pause 200
    endif
  next C%
  
' ringing tones of the world, uncomment your preferred ringing tone!
  print
  pause 3000
  for C%=1 to (rnd * 4) + 1
' North America
    play tone 440,480,2000
    pause 6000
' Europe
    'play tone 425,425,1000
    'pause 4000
' France
    'play tone 440,440,1000
    'pause 4000
' UK, Ireland, New Zealand, Australia (gday my cmm2 mates!)
    'play tone 400,450,400
    'pause 600
    'play tone 400,450,400
    'pause 2600
' Japan
    'play tone 400,420,1000
    'pause 3000
  next C%
  play tone 2100,2100,3000
  pause 2500 
end sub

function getchar$() as string
  altflag% = 0
  winflag% = 0
  getchar$ = ""
  if keydown(7) > 0 then
    kf% = keydown(7)
    if (kf% AND &b00000001) = &b00000001 then altflag% = 1 
    if (kf% AND &b00010000) = &b00010000 then altflag% = 1
    if (kf% AND &b00000100) = &b00000100 then winflag% = 1
    if (kf% AND &b00100000) = &b01000000 then winflag% = 1
  end if
  kp% = keydown(0)
  if kp% > 0 then
    kp% = keydown(1)
    if kl% <> kp% then
      getchar$ = chr$(kp%)    
      timer = 0
    else
      if timer > kreprate then
        getchar$ = chr$(kp%)
        timer = 0
      end if
    end if
    kl% = kp%
    pause 10
  end if
end function

sub printstatus(str$)
  local x%
  blit fwidth%*(xoffset%+1), (height%+14)*fheight%, fwidth%*(xoffset%+1), (height%+13)*fheight%, fwidth%*width, 2*fheight%
  for x% = 0 to 77
    text fwidth%*(xoffset%+1+x%), (height%+15)*fheight%, " "
  next x%
  text fwidth%*(xoffset%+1), (height%+15)*fheight%, str$
  updateonline
end sub

sub blinkcursor
  local x%, y%
  x% = xpos%
  y% = ypos%
  select case cursortype%
  case 0
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) " ";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "_";
  case 1
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "/";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "-";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "\";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "|";
  case 2
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) ":";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "-";
  case 3
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "<";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "^";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) ">";
    print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) "v";
  end select
  print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) scr_text$(x%,y%);
end sub

sub clearscreen
local x%, y%
  for y% = 0 to height-1
    for x% = 0 to width-1
      scr_text$(x%,y%) = " "    ' space
    next x%
  next y%
  xpos% = 0 : ypos% = 0
  updatescreen
end sub

sub scrollup
local x%, y%
  for y% = 1 to height-1
    for x% = 0 to width-1
      scr_text$(x%,y%-1) = scr_text$(x%,y%)
    next x%
  next y%
  for x% = 0 to width-1
    scr_text$(x%,height-1) = " "
  next x%
  blit fwidth%*xoffset%, (fheight%*2)+(fheight%*(yoffset%-1)), fwidth%*xoffset%, fheight%+(fheight%*(yoffset%-1)), fwidth%*(width), fheight%*(height-1)
  for x% = 0 to width-1
    print @((x%+xoffset%)*fwidth%,(height-1+yoffset%)*fheight%) scr_text$(x%,height-1);
  next x%
end sub

sub updatescreen
local x%, y%
  for y% = 0 to height-1
    for x% = 0 to width-1
      print @((x%+xoffset%)*fwidth%,(y%+yoffset%)*fheight%) scr_text$(x%,y%);
    next x%
  next y%
end sub

sub updatechar
  print @((xpos%+xoffset%)*fwidth%,(ypos%+yoffset%)*fheight%) scr_text$(xpos%,ypos%);
end sub

sub setbuffer
local x%, y%
  for y% = 0 to height-1
    for x% = 0 to width-1
      scr_text$(x%,y%) = " "    ' space
      scr_color%(x%,y%) = 1     ' white
    next x%
  next y%
end sub

sub printc(str$)
  select case str$
  ' carriage return
  case chr$(10)
    str$ = chr$(13)
  ' linefeed
  case chr$(13)
    xpos% = 0
    ypos% = ypos% + 1
    if ypos% > height-1 then 
      ypos% = height-1
      scrollup
    end if
  ' backspace, destructive
  case chr$(8)
    xpos% = xpos% - 1
    if xpos% < 0 then
      xpos% = 0
      ypos% = ypos% - 1
      if ypos% < 0 then ypos% = 0
    end if
    scr_text$(xpos%, ypos%) = " "
  ' delete, destructive
  case chr$(127)
    scr_text$(xpos%, ypos%) = " "  
  ' left arrow, non-destructive
  case chr$(130)
    xpos% = xpos% - 1
    if xpos% < 0 then
      xpos% = 79
      ypos% = ypos% - 1
      if ypos% < 0 then ypos% = 0 : xpos% = 0
    end if
  ' right arrow, non-destructive
  case chr$(131)
    xpos% = xpos% + 1
    if xpos% > 79 then
      xpos% = 0
      ypos% = ypos% + 1
      if ypos% > 23 then ypos% = 23 : xpos% = 79
    end if
  ' up arrow
  case chr$(128)
  ' down arrow
  case chr$(129)
  ' escape
  case chr$(27)
  ' tav
  case chr$(9)
    end sub
  ' everything else
  case else
    scr_text$(xpos%, ypos%) = left$(str$,1)
    updatechar
    xpos% = xpos% + 1
    if xpos% > width-1 then 
      xpos% = 0
      ypos% = ypos% + 1
      if ypos% > height-1 then 
        ypos% = height-1
        scrollup
      end if
    end if
  end select
end sub

sub printsret(str$)
local x%
  for x% = 1 to len(str$)
    printc(mid$(str$,x%,1))
  next x%
  printc(chr$(13))
end sub

sub prints(str$)
local x%
  for x% = 1 to len(str$)
    printc(mid$(str$,x%,1))
  next x%
end sub

sub printxy(x1%,y1%,str$)
local x%
  if x1% < width and y1% < height then
    xpos% = x1%
    ypos% = y1%
    for x% = 1 to len(str$)
      printc(mid$(str$,x%,1))
    next x%
  end if
end sub
